# Run dependency injections
import os
import tekleo_common_utils
from injectable import load_injection_container
load_injection_container(str(os.path.dirname(tekleo_common_utils.__file__)))
load_injection_container('../../')
from tekleo_common_utils_ai.utils_dataset_google import UtilsDatasetGoogle
from tekleo_common_utils_ai.utils_dataset_labelme import UtilsDatasetLabelme
from tekleo_common_utils_ai.utils_dataset_coco import UtilsDatasetCoco

utils_dataset_google = UtilsDatasetGoogle()
utils_dataset_labelme = UtilsDatasetLabelme()
utils_dataset_coco = UtilsDatasetCoco()

json_credentials_filepath = 'tab32-xray-vision-7d4fd8415bc7.json'
data_filepath = 'data-00001-of-00001.jsonl'
google_od_samples = utils_dataset_google.load_samples_from_jsonl(json_credentials_filepath, data_filepath)

labelme_folderpath = "/Users/leo/tekleo/tekleo-common-utils-ai/tekleo_common_utils_ai/test/dataset_google/labelme_conversion"
utils_dataset_labelme.save_samples_to_folder(google_od_samples, labelme_folderpath)

coco_folderpath = "/Users/leo/tekleo/tekleo-common-utils-ai/tekleo_common_utils_ai/test/dataset_google/coco_conversion"
utils_dataset_coco.save_samples_to_folder(google_od_samples, coco_folderpath)


import os
from typing import List, Dict, Tuple, Optional
import concurrent.futures
from itertools import repeat
from injectable import injectable, autowired, Autowired
from simplestr import gen_str_repr_eq
from pydantic import BaseModel
from tekleo_common_message_protocol import OdSample, OdLabeledBox, RectangleRelative
from tekleo_common_utils import UtilsImage


# Labelme JSON models
#-----------------------------------------------------------------------------------------------------------------------
@gen_str_repr_eq
class LabelmeShape(BaseModel):
    label: str
    points: List[Tuple[float, float]]
    group_id: Optional[str]
    shape_type: str
    flags: Dict[str, str]

    def __init__(self, label: str, points: List[Tuple[float, float]], group_id: str, shape_type: str, flags: Dict[str, str]) -> None:
        super().__init__(label=label, points=points, group_id=group_id, shape_type=shape_type, flags=flags)


@gen_str_repr_eq
class LabelmeSample(BaseModel):
    version: str
    flags: Dict[str, str]
    shapes: List[LabelmeShape]
    imagePath: str
    imageData: str
    imageHeight: int
    imageWidth: int

    def __init__(self, version: str, flags: Dict[str, str], shapes: List[LabelmeShape], imagePath: str, imageData: str, imageHeight: int, imageWidth: int) -> None:
        super().__init__(version=version, flags=flags, shapes=shapes, imagePath=imagePath, imageData=imageData, imageHeight=imageHeight, imageWidth=imageWidth)
#-----------------------------------------------------------------------------------------------------------------------


@injectable
class UtilsDatasetLabelme:
    @autowired
    def __init__(self, utils_image: Autowired(UtilsImage)):
        self.utils_image = utils_image

    def load_sample_from_image_and_labelme_json(self, image_file_path: str, json_file_path: str) -> OdSample:
        pass

    def save_sample_to_folder(self, od_sample: OdSample, folder_path: str) -> bool:
        # Basic image info
        image_height = od_sample.image.height
        image_width = od_sample.image.width
        image_file_path = folder_path + '/' + od_sample.name + '.jpg'
        json_file_path = folder_path + '/' + od_sample.name + '.json'

        # Convert all boxes
        shapes = []
        for box in od_sample.boxes:
            shape = LabelmeShape(
                label=box.label,
                points=[
                    (box.region.x * image_width, box.region.y * image_height),
                    ((box.region.x + box.region.w) * image_width, (box.region.y + box.region.h) * image_height)
                ],
                group_id=None,
                shape_type="rectangle",
                flags={}
            )

            shapes.append(shape)

        # Convert the sample
        labelme_sample = LabelmeSample(
            version="5.0.1",
            flags={},
            shapes=shapes,
            imagePath=image_file_path.split('/')[-1],
            imageData=self.utils_image.encode_image_pil_as_base64(od_sample.image),
            imageHeight=image_height,
            imageWidth=image_width,
        )

        # Save image file
        self.utils_image.save_image_pil(od_sample.image, image_file_path)

        # Save json file
        json_str = str(labelme_sample.json())
        json_file = open(json_file_path, "w")
        json_file.write(json_str)
        json_file.close()

        return True

    def save_samples_to_folder(self, od_samples: List[OdSample], folder_path: str) -> bool:
        results = []
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=8)
        for result in executor.map(self.save_sample_to_folder, od_samples, repeat(folder_path)):
            results.append(result)
        return True



from .utils_dataset_coco import UtilsDatasetCoco
from .utils_dataset_google import UtilsDatasetGoogle
from .utils_dataset_labelme import UtilsDatasetLabelme
from .utils_dataset_pascalvoc import UtilsDatasetPascalvoc
from .utils_detectron_model import UtilsDetectronModel
from .utils_visualize_od import UtilsVisualizeOd

__all__ = [
    UtilsDatasetCoco,
    UtilsDatasetGoogle,
    UtilsDatasetLabelme,
    UtilsDatasetPascalvoc,
    UtilsDetectronModel,
    UtilsVisualizeOd
]



import os
import datetime
import collections
import json
import uuid
import imgviz
import labelme
import numpy
import pycocotools.mask
from typing import List, Dict, Tuple, Optional
import concurrent.futures
from itertools import repeat
from injectable import injectable, autowired, Autowired
from simplestr import gen_str_repr_eq
from pydantic import BaseModel
from tekleo_common_message_protocol import OdSample, OdLabeledBox, RectangleRelative
from tekleo_common_utils import UtilsImage


@injectable
class UtilsDatasetCoco:
    @autowired
    def __init__(self, utils_image: Autowired(UtilsImage)):
        self.utils_image = utils_image

    def _build_default_class_labels(self, od_samples: List[OdSample]) -> List[str]:
        class_labels = []
        for s in od_samples:
            for b in s.boxes:
                class_labels.append(b.label)
        class_labels = set(class_labels)
        class_labels = list(class_labels)
        class_labels = sorted(class_labels)
        adjusted_class_labels = []
        adjusted_class_labels.append("__ignore__")
        adjusted_class_labels.extend(class_labels)
        return adjusted_class_labels

    # class_labels must have __ignore__ class as first item
    def save_samples_to_folder(self, od_samples: List[OdSample], folder_path: str, class_labels: List[str] = []) -> bool:
        # Fill class labels if default
        if class_labels is None or len(class_labels) == 0:
            class_labels = self._build_default_class_labels(od_samples)
        print('UtilsDatasetCoco.save_samples_to_folder(): class_labels=' + str(class_labels))

        # Make sure nested dirs exits
        if not os.path.exists(folder_path + "/JPEGImages"):
            os.makedirs(folder_path + "/JPEGImages")
        if not os.path.exists(folder_path + "/Visualization"):
            os.makedirs(folder_path + "/Visualization")

        data = dict(
            info=dict(
                description=None,
                url=None,
                version=None,
                year=datetime.datetime.now().year,
                contributor=None,
                date_created=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            ),
            licenses=[dict(url=None, id=0, name=None,)],
            images=[
                # license, url, file_name, height, width, date_captured, id
            ],
            type="instances",
            annotations=[
                # segmentation, area, iscrowd, image_id, bbox, category_id, id
            ],
            categories=[
                # supercategory, id, name
            ],
        )

        class_name_to_id = {}
        for i in range(0, len(class_labels)):
            class_id = i - 1  # starts with -1
            class_name = class_labels[i]
            if class_id == -1:
                assert class_name == "__ignore__"
                continue
            class_name_to_id[class_name] = class_id
            data["categories"].append(
                dict(supercategory=None, id=class_id, name=class_name,)
            )

        out_ann_file = folder_path + "/annotations.json"
        for od_sample_index in range(0, len(od_samples)):
            od_sample = od_samples[od_sample_index]
            image_width = od_sample.image.width
            image_height = od_sample.image.height
            print("UtilsDatasetCoco.save_samples_to_folder(): Generating coco dataset from:", od_sample.name)

            out_img_file = folder_path + "/JPEGImages/" + od_sample.name + ".jpg"
            self.utils_image.save_image_pil(od_sample.image, out_img_file)
            image_cv = self.utils_image.convert_image_pil_to_image_cv(od_sample.image)

            data["images"].append(
                dict(
                    license=0,
                    url=None,
                    file_name="JPEGImages/" + od_sample.name + ".jpg",
                    height=image_height,
                    width=image_width,
                    date_captured=None,
                    id=od_sample_index,
                )
            )

            masks = {}  # for area
            segmentations = collections.defaultdict(list)  # for segmentation
            for box in od_sample.boxes:
                points = [(box.region.x * image_width, box.region.y * image_height), ((box.region.x + box.region.w) * image_width, (box.region.y + box.region.h) * image_height)]
                label = box.label
                group_id = None
                shape_type = "rectangle"
                mask = labelme.utils.shape_to_mask(
                    image_cv.shape[:2], points, shape_type
                )

                if group_id is None:
                    group_id = uuid.uuid1()

                instance = (label, group_id)

                if instance in masks:
                    masks[instance] = masks[instance] | mask
                else:
                    masks[instance] = mask

                if shape_type == "rectangle":
                    (x1, y1), (x2, y2) = points
                    x1, x2 = sorted([x1, x2])
                    y1, y2 = sorted([y1, y2])
                    points = [x1, y1, x2, y1, x2, y2, x1, y2]
                if shape_type == "circle":
                    (x1, y1), (x2, y2) = points
                    r = numpy.linalg.norm([x2 - x1, y2 - y1])
                    # r(1-cos(a/2))<x, a=2*pi/N => N>pi/arccos(1-x/r)
                    # x: tolerance of the gap between the arc and the line segment
                    n_points_circle = max(int(numpy.pi / numpy.arccos(1 - 1 / r)), 12)
                    i = numpy.arange(n_points_circle)
                    x = x1 + r * numpy.sin(2 * numpy.pi / n_points_circle * i)
                    y = y1 + r * numpy.cos(2 * numpy.pi / n_points_circle * i)
                    points = numpy.stack((x, y), axis=1).flatten().tolist()
                else:
                    points = numpy.asarray(points).flatten().tolist()

                segmentations[instance].append(points)
            segmentations = dict(segmentations)

            for instance, mask in masks.items():
                cls_name, group_id = instance
                if cls_name not in class_name_to_id:
                    continue
                cls_id = class_name_to_id[cls_name]

                mask = numpy.asfortranarray(mask.astype(numpy.uint8))
                mask = pycocotools.mask.encode(mask)
                area = float(pycocotools.mask.area(mask))
                bbox = pycocotools.mask.toBbox(mask).flatten().tolist()

                data["annotations"].append(
                    dict(
                        id=len(data["annotations"]),
                        image_id=od_sample_index,
                        category_id=cls_id,
                        segmentation=segmentations[instance],
                        area=area,
                        bbox=bbox,
                        iscrowd=0,
                    )
                )

            viz = image_cv
            if masks:
                labels, captions, masks = zip(
                    *[
                        (class_name_to_id[cnm], cnm, msk)
                        for (cnm, gid), msk in masks.items()
                        if cnm in class_name_to_id
                    ]
                )
                viz = imgviz.instances2rgb(
                    image=image_cv,
                    labels=labels,
                    masks=masks,
                    captions=captions,
                    font_size=15,
                    line_width=2,
                )
            out_viz_file = folder_path + "/Visualization/" + od_sample.name + ".jpg"
            self.utils_image.save_image_cv(viz, out_viz_file)

        with open(out_ann_file, "w") as f:
            json.dump(data, f)

        return True



import json
import os
import tempfile
from typing import List
import concurrent.futures
from itertools import repeat
from google.cloud import storage
from injectable import injectable, autowired, Autowired
from tekleo_common_message_protocol import OdSample, OdLabeledBox, RectangleRelative
from tekleo_common_utils import UtilsImage


@injectable
class UtilsDatasetGoogle:
    @autowired
    def __init__(self, utils_image: Autowired(UtilsImage)):
        self.utils_image = utils_image

    def load_sample_from_jsonl_line(self, json_credentials_filepath: str, jsonl_data_line: str) -> OdSample:
        # Parse JSON object
        json_dict = json.loads(jsonl_data_line)
        image_gcs_uri = str(json_dict['imageGcsUri'])
        bounding_box_annotations = json_dict['boundingBoxAnnotations']

        # Create name, file name, and temp path
        od_sample_name = image_gcs_uri.split('/')[-1].split('.')[0]
        image_file_name = image_gcs_uri.split('/')[-1]
        image_file_path = tempfile.gettempdir() + '/' + image_file_name
        # print('UtilsDatasetGoogle.load_sample_from_jsonl_line(): od_sample_name=' + str(od_sample_name) + ', image_file_name=' + str(image_file_name) + ', image_file_path=' + str(image_file_path))

        # Download the image, create client, enter a bucket, select blob and download it
        storage_client = storage.Client.from_service_account_json(json_credentials_filepath)
        bucket = storage_client.bucket("ml_image_stage")
        blob_name = image_gcs_uri.replace('gs://ml_image_stage/', '')
        blob = bucket.blob(blob_name)
        blob.download_to_filename(image_file_path)

        # Open the image
        image_pil = self.utils_image.open_image_pil(image_file_path)

        # Parse boxes
        od_sample_boxes = []
        for bounding_box_annotation in bounding_box_annotations:
            x = bounding_box_annotation['xMin']
            y = bounding_box_annotation['yMin']
            w = bounding_box_annotation['xMax'] - x
            h = bounding_box_annotation['yMax'] - y
            od_labeled_box_region = RectangleRelative(x, y, w, h)
            od_labeled_box_label = bounding_box_annotation['displayName']
            od_labeled_box = OdLabeledBox(od_labeled_box_label, od_labeled_box_region)
            od_sample_boxes.append(od_labeled_box)

        # Create finalize sample
        return OdSample(od_sample_name, image_pil, od_sample_boxes)

    def load_samples_from_jsonl(self, json_credentials_filepath: str, jsonl_data_file_path: str) -> List[OdSample]:
        # Read JSON file
        jsonl_data_file = open(jsonl_data_file_path, 'r')
        jsonl_data_text = jsonl_data_file.read()
        jsonl_data_file.close()
        jsonl_data_lines = jsonl_data_text.split("\n")
        jsonl_data_lines = [l.strip() for l in jsonl_data_lines if len(l.strip()) > 0]

        # Initialize list
        od_samples = []

        # Load samples in parallel
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=8)
        for od_sample in executor.map(self.load_sample_from_jsonl_line, repeat(json_credentials_filepath), jsonl_data_lines):
            od_samples.append(od_sample)

        # Return results
        return od_samples



from typing import List
import cv2
import numpy
from numpy import ndarray
from tekleo_common_message_protocol import OdPrediction
from tekleo_common_utils import UtilsImage
from injectable import injectable, autowired, Autowired


@injectable
class UtilsVisualizeOd:
    @autowired
    def __init__(self, utils_image: Autowired(UtilsImage)):
        self.utils_image = utils_image

    def debug_predictions(self, image_cv: ndarray, predictions: List[OdPrediction], class_labels: List[str]):
        result_image_cv = image_cv.copy()
        colors = [
            (0, 255, 0),
            (0, 0, 255),
            (255, 0, 0),
            (238, 48, 167),
            (205, 186, 150),
            (255, 193, 193)
        ]

        for prediction in predictions:
            # Determine color
            color_index = class_labels.index(prediction.label)
            color = colors[color_index]

            # Draw region box
            box_x1 = prediction.region.x
            box_x2 = prediction.region.x + prediction.region.w
            box_y1 = prediction.region.y
            box_y2 = prediction.region.y + prediction.region.h
            result_image_cv = cv2.rectangle(result_image_cv, (box_x1, box_y1), (box_x2, box_y2), color, 1)

            # Draw label
            #result_image_cv = cv2.putText(result_image_cv, prediction.label, (prediction.region.x, prediction.region.y), cv2.FONT_HERSHEY_SIMPLEX, 0.4, color, 1,  cv2.LINE_AA)

            # Draw mask polygon
            polygon_x_values = [point.x for point in prediction.mask]
            polygon_y_values = [point.y for point in prediction.mask]
            polygon_array = [(x, y) for x, y in zip(polygon_x_values, polygon_y_values)]
            vertices = numpy.array(polygon_array)
            mask_background = numpy.zeros(shape=image_cv.shape, dtype=numpy.uint8)
            mask = cv2.fillPoly(mask_background, [vertices], color)
            result_image_cv = cv2.addWeighted(result_image_cv, 1, mask, 0.4, 0)

        self.utils_image.debug_image_cv(result_image_cv)
